	page	60,132
        Title   KEY_PIO.ASM Matrix keypad driver
        Subttl  Copyright JK microsystems 1997-2004

;************************************************************************;
;*								       	*;
;*      Copyright JK microsystems 1997-2004  -  All Rights Reserved     *;
;*									*;
;*	This software is NOT shareware, freeware, or public domain.	*;
;*	It is the property of JK microsystems.				*;
;*									*;
;*	Customers of JK microsystems may modify the source code		*; 
;*	and/or distribute the binary image of this software without 	*;
;*	additional costs provided it is run only on hardware 		*;
;*	manufactured by JK microsystems.  All other use is expressly 	*;
;*	prohibited.							*;
;*									*;
;*	Update Log							*;
;*									*;
;*	Version	Date		Comments		Progammer	*;
;*	-------	-------		----------------------	----------	*;
;*	1.0	06-10-97	Initial release		jds		*;
;*	2.0	12-17-97	modified KEPAD.ASM			*;
;*				for 386ex		jds		*;
;*	2.1	02-12-99        Fixed problem with bios			*;
;*				32d and later				*;
;*							jds		*;
;*      3.0     06-23-00        modified KEYPADED.ASM                   *;
;*                              for LogicFlex           ew              *;
;*                                                                      *;
;*      4.0     08-20-02        modified KEYPADLF.ASM                   *;
;*                              for Flashlite 186       ew              *;
;*									*;
;*	5.0	04-09-03	modified 186 version 			*;
;*				to KEYPADPF.ASM for			*;
;*				Picoflash		jds		*;
;*									*;
;*	5.0	04-09-03	modified KEYPADPF.ASM 			*;
;*				to KEY_PIO.ASM for			*;
;*				Picoflash/PicoI/O	jds		*;
;************************************************************************;

Code_seg	Segment
	assume	cs:code_seg,ds:code_seg,ss:code_seg

	org	100h
start:					; Put the initialization code at
	jmp	init			; the end so we don't have to save it
	

;-----------------------------------------------------------------------------
;
;	This is invoked with each bios int16h keyboard interrupt
;
key_int:
	cmp	cs:[state],1		; is there a char pending
	jne	key_int_done

	cmp	ah,0			; function 0?
	jne	check_func_1		; no, check 1

	mov	ah,0			; zero the scan code
	mov	al,cs:[buffer]		; get the char in al
	mov	cs:[state],2		; set state to check for key up
	iret				; and return		

check_func_1:
	cmp	ah,1			; function 1?
	jne	key_int_done		; no, give the old handler a shot

	cmp	al,al			; set the zero flag
	mov	ah,0			; zero the scan code
	mov 	al,cs:[buffer]		; get the char, don't change state
	iret


key_int_done:	
	db	0EAh			; far jump, masm too stupid to do it
key_off	dw	0			; to old key isr
key_seg	dw	0		


;-----------------------------------------------------------------------------
;
;	This is invoked with each bios tick interrupt
;
tick_int:
	push	ax			; save the context
	push	bx		
	push	si
	push	di
	push	ds

	cmp	cs:[state],0		; idle
	je	state0
	cmp	cs:[state],1		; char ready, waiting for int16
	je	state1
	cmp	cs:[state],2		; key down, waiting for up
	je	state2		

	mov	cs:[state],0		; unknown state, make it idle	

tick_int_done:
	pop	ds
	pop	di
	pop	si
	pop	bx
	pop	ax

	 db	0EAh			; far jump
tick_off dw	0			; to old tick isr
tick_seg dw	0


;-----------------------------------------------------------------------------
;
;	State 0 - Idle, scan the keypad
;
;	If key pressed, put it in buffer and set state to 1
;	If not pressed, stay in state 0
;
state0:
	call	scan			; scan the keypad
	jc	tick_int_done		; carry set, no keypress
	mov	cs:[buffer],al		; put the char in the buffer
	mov	cs:[state],1		; set state 1
	jmp	tick_int_done		; and leave


;-----------------------------------------------------------------------------
;
;	State 1 - Char in buffer, waiting for int16 to get it
;
state1:
	jmp	tick_int_done		; nothing to do here


;-----------------------------------------------------------------------------
;
;	State 2 - Waiting for key to be unpressed
;
state2:
	call	scan			; scan the keypad
	jnc	tick_int_done		; key is down, stay in state
	mov	cs:[state],0		; key up, set state 0
	jmp	tick_int_done		; done

;-----------------------------------------------------------------------------
;
;	Scan the keypad, return row and column in BL and BH respectively.
;
;	Carry set if key up, carry cleared if key down
;

;	Before we can change any of the row output bits we have
;	to get the output shadow register from the bios data area

Scan:
	push	dx
	push	ds			; area 
	mov	ax,40h			; so we can get the out register
	mov	ds,ax
	mov	ax,ds:[0B0h]		; get the lower word 
	mov	cs:[lo_word],ax		; store it locally
	mov	al,ds:[0B2h]		; get the upper byte
	mov	cs:[hi_byte],al		; store it locally
	mov	al,ds:[0B3h]		; get busy flag in AL
	pop	ds			; restore registers
	cmp	al,0			; is the busy flag set?
	je	scan_ok			; no, ok to scan
	jmp	no_key_pressed		; yes, can't scan now, leave

scan_ok:
;	Got the output shadow register locally. Now we can scan
;	The scan routine is all new, rewritten to save time.

	mov	dx,606h			; set port PA to outputs
	in	al,dx			; must assume someone else might
	or	al,1			; have changed it on us
	out	dx,al
	mov	dx,600h			; point to input port PH
	mov	al,7
	out	dx,al
	mov	dx,601h			; ready to read port PH	

	and	cs:[lo_word],word ptr 0FFF0h ; clear all 4 row bits
	call	load_chip		; write it out	
	in	al,dx			; read the columns
	cmp	al,0Fh			; see if all columns are high
	jne	do_scan			; no, a key is pressed
	jmp	no_key_pressed		; yes, end it now, no key pressed

do_scan:
	mov	bx,0			; clear it so we can return data
	or	cs:[lo_word],word ptr 0Eh    ; set all 4 bits
	and	cs:[lo_word],word ptr 0FFFEh ; clear row zero
	call	load_chip		; write it out	
	in	al,dx			; read the columns
	cmp	al,0Fh			; if all high, goto next row
	je	row_1
	jmp	decode_column		
row_1:
	add	bl,4			; bump row count
	or	cs:[lo_word],word ptr 0Eh    ; set all 4 bits
	and	cs:[lo_word],word ptr 0FFFDh ; clear row one
	call	load_chip		; write it out	
	in	al,dx			; read the columns
	cmp	al,0Fh			; if all high, goto next row
	je	row_2
	jmp	decode_column		
row_2:
	add	bl,4			; bump row count
	or	cs:[lo_word],word ptr 0Eh    ; set all 4 bits
	and	cs:[lo_word],word ptr 0FFFBh ; clear row 2
	call	load_chip		; write it out	
	in	al,dx			; read the columns
	cmp	al,0Fh			; if all high, goto next row
	je	row_3
	jmp	decode_column		
row_3:
	add	bl,4			; bump row count
	or	cs:[lo_word],word ptr 0Eh    ; set all 4 bits
	and	cs:[lo_word],word ptr 0FFF7h ; clear row 3
	call	load_chip		; write it out	
	in	al,dx			; read the columns
	cmp	al,0Fh			; if all high, goto next row
	je	no_key_pressed		; should never get here

decode_column:		
	cmp	al,0Eh			; col 0 pressed?
	je	key_pressed
	inc	bl			; no, bump col count
	cmp	al,0Dh
	je	key_pressed
	inc	bl
	cmp	al,0Bh
	je	key_pressed
	inc	bl
	cmp	al,07h
	je	key_pressed
	jmp	no_key_pressed		; should never get here	

Key_pressed:
	push	di
	mov	di,offset keymap
	mov	al,cs:[di+bx]
	pop	di	
	pop	dx
	clc
	ret

no_key_pressed:
	pop	dx
	stc
	ret	

;-----------------------------------------------------------------------------
;
;	Load Chip - local routine to move hi_byte and lo_word to
;	NJU3718 chip.

load_chip:
	push	ax
	push	bx
	push	cx
	push	dx

	mov	bl,cs:[hi_byte]

	shl	bl,1			; left-justify them
	shl	bl,1
	shl	bl,1
	shl	bl,1

	mov	dx,604h			; point to I/O port
	mov	al,0Ch			; CLR/ STB/ CLK DATA
	out	dx,al			;  1    1    0   0
	mov	cx,4
hi:
	rcl	bl,1			; get current bit into carry
	rcl	al,1			; get carry into al, bit 0
	and	al,1			; clear all but bit 0
	or	al,0Eh			; set CLR/ STB/ CLK
	out	dx,al			; send it
	and	al,0FDh			; clear CLK
	out	dx,al			; send it
	loop	hi			; loop through rest

	mov	bx,cs:[lo_word]		

	mov	cx,16			; do all 16 bits
lo:
	rcl	bx,1			; get current bit into carry
	rcl	al,1			; get carry into al, bit 0
	and	al,1			; clear all but bit 0
	or	al,0Eh			; set CLR/ STB/ CLK
	out	dx,al			; send it
	and	al,0FDh			; clear CLK
	out	dx,al			; send it
	loop	lo			; loop through rest

	mov	al,8
	out	dx,al			; strobe the data to the outputs
	mov	al,0Ch			; make STB/ high
	out	dx,al

	pop	dx
	pop	cx
	pop	bx
	pop	ax

	ret	

;-----------------------------------------------------------------------------
;
;	Data Area
;
state	db	0			
buffer	db	0			; key code 

lo_word		dw	?
hi_byte 	db	?

keymap:
        db                      '1','2','3','A'         ; 64-0001
        db                      '4','5','6','B'         ; 64-0001
        db                      '7','8','9','C'         ; 64-0001 
        db                      '*','0','#','D'         ; 64-0001          

;-----------------------------------------------------------------------------
;
;	This is our initialization code
;
init:
        mov     dx,606h                 ; point to Port Config Reg
        in      al,dx                   ; get current contents
        or      al,08h                  ; set Port D as output
        out     dx,al                   ; do it

	mov	ax,0			; point to interrupt vector table
	mov 	ds,ax

	cli				; no ints while changing vectors

	mov	ax,ds:[16h * 4]		; get int 16h offset
	mov	cs:[key_off],ax		; save it and replace
	mov	ds:[16h * 4],offset key_int 

	mov	ax,ds:[16h * 4 + 2]	; get int 16h segment
	mov	cs:[key_seg],ax		; save it
	mov	ax,cs			; get our segment
	mov	ds:[16h * 4 + 2],ax	; and install it

	mov	ax,ds:[1Ch * 4]		; get int 1Ch offset
	mov	cs:[tick_off],ax	; save it and replace
	mov	ds:[1Ch * 4],offset tick_int 

	mov	ax,ds:[1Ch * 4 + 2]	; get int 1Ch segment
	mov	cs:[tick_seg],ax	; save it
	mov	ax,cs			; get our segment
	mov	ds:[1Ch * 4 + 2],ax	; and install it

	sti				; ints back on

	mov	ax,cs			; put ds back
	mov	ds,ax

	mov	ax,3100h		; TSR function code
	mov 	dx,offset init	 	; get last address used
	mov 	cl,4			; convert to paragraphs
	shr     dx,cl			; by dividing by 16
	inc	dx			; and rounding up by one

	int	21h			; tell dos to do it

code_seg 	ends

	end	start

