start:
	open "o",1,"cons:"		' open console out

	' Array of thermistor resistance values in K ohms starting 
	' at -40 C and incrementing in 5 degree C steps
					
	dim t(0 to 33)			
	t(0)=4397:   t(1)=3088:   t(2)=2197:   t(3)=1581:   t(4)=1151
	t(5)=846.6:  t(6)=629.0:  t(7)=471.6:  t(8)=357.0:  t(9)=272.5
	t(10)=209.7: t(11)=162.7: t(12)=127.1: t(13)=100.0: t(14)=79.22
	t(15)=63.17: t(16)=50.68: t(17)=40.90: t(18)=33.19: t(19)=27.09
	t(20)=22.22: t(21)=18.32: t(22)=15.18: t(23)=12.64: t(24)=10.57
	t(25)=8.873: t(26)=7.481: t(27)=6.337: t(28)=5.383: t(29)=4.594
	t(30)=3.934: t(31)=3.380: t(32)=2.915: t(33)=2.522	

	a% = 0	
	call InitIO(a%)
	call GetVersion(a%)
	print #1,"Driver version is ";
	print #1, using "#.#";(a%/10)

menuloop:	
	print #1,chr$(27);"[2J";	' clear screen, home cursor
	print #1,chr$(27);"[H";		
	print #1,"PicoFlash PicoIO Test Program"
	print #1,"  Press I to test inputs"
	print #1,"  Press O to test outputs"
	print #1,"  Press A to test A/D"
	print #1,"  Press T to show thermistor temperature"
	print #1,"  Press Q to quit"
	print #1,">";
	a$ = ""
	while a$ = ""
		a$ = inkey$
	wend
	if a$ = "i" then gosub inputtest
	if a$ = "o" then gosub outputtest
	if a$ = "a" then gosub adtest
	if a$ = "t" then gosub ShowTemp
	if a$ = "q" then goto quit
	goto menuloop
		
outputtest:
	print #1,""
	print #1,"Enter S to set point, C to clear point,"
	print #1,">";
	linein$ = ""
	gosub linein
	if lcase$(linein$) = "q" then return
	if lcase$(linein$) = "s" then gosub setpoint
	if lcase$(linein$) = "c" then gosub clearpoint
	goto outputtest

setpoint:
	print #1,"Enter point to set, 0-19, q to quit ";
	gosub linein
	if linein$ = "q" then return
	a% = val(linein$)
	call SetIOPt(a%)
	goto setpoint

clearpoint:
	print #1,"Enter point to clear, 0-19, q to quit ";
	gosub linein
	if linein$ = "q" then return
	a% = val(linein$)
	call ClrIOPt(a%)
	goto clearpoint

inputtest:
	print #1,""
	print #1,"Enter point to get, 0-31, q to quit ";
	gosub linein
	if linein$ = "q" then return
	a% = val(linein$)
	call GetIOPt(a%)
	print #1,a%
	goto inputtest

adtest:	
	print #1,"Enter A/D channel to read, 0-10, q to quit ";
	gosub linein
	if linein$ = "q" then return
getadloop:
	if inkey$ = "q" then return
	a% = val(linein$)
	call GetAD(a%)
	v = (a% / 819)
	print #1,"Volts =";
	print #1, using "#.###";v;
	print #1,"   A/D Count =";a%
	goto getadloop


quit:	
	print #1,chr$(27);"[2J";	' clear screen, home cursor
	print #1,chr$(27);"[H";		
	end



linein:
	linein$ = ""
linemore:
	a$ = INKEY$
	IF a$ = "" THEN GOTO linemore
	IF a$ = CHR$(13) THEN GOTO linedone
	IF a$ <> CHR$(8) THEN GOTO getchar
	PRINT #1, CHR$(8); CHR$(32); CHR$(8);
	linein$=left$(linein$,(len(linein$)-1))
	GOTO linemore
getchar:    
	PRINT #1, a$;
	linein$ = linein$ + a$
	GOTO linemore
linedone:
	PRINT #1, ""
	RETURN


ShowTemp:
	print #1,chr$(27);"[2J";	' clear screen, home cursor
ShowTempLoop:
	print #1,chr$(27);"[H";		
		
	a%=10				' get channel A/D channel 10
	call GetAD(a%)
	ADcounts=a%	

	Vin=ADcounts/1000		' Vin is A/D volts in
	I=Vin/82500			' I is current through resistors
	Vtc=4.096-(I*18220+Vin)		' Vtc is volts across thermistor
	Rtc=(Vtc/I)/1000		' Rtc is measured thermistor K ohms

	i=0				' point to start of array
	Th=-40				' first element is for -40c
	while t(i)>Rtc			' step through array until element
		i=i+1			' is greater than measured resistance
		Th=Th+5			' increment temperature with each step
	wend

	Trh=t(i)			' Trh is closest larger resistance
	Trl=t(i-1)			' Trl is closest smaller resistance
	TrDiff=(5/(Trl-Trh))		' take the difference and divide by 5
	Tcorr=(trh-Rtc)*TrDiff		' take the difference between the
					' larger and actual values and scale it
	Tc=Th+Tcorr			' Apply it to temperature 
	Tf=((Tc*9)/5)+32		' Calculate degrees F

	print #1," Vtc      Rtc     Tc     Tf"
	print #1,"-----    -----   ----   ----"

	print #1,using "#.###";Vtc;	' Show A/D volts accross thermistor
	print #1,"  ";	
	print #1,using "####.##";Rtc;	' Show calculated thermistor ohms
	print #1,"  ";
	print #1,using "###.#";Tc;	' Show temperature in degrees C
	print #1,"  ";
	print #1,using "###.#";Tf	' Show temperature in degrees F

	print #1,""
	print #1,"Vtc = Volts accross thermistor"
	print #1,"Rtc = Calculated thermistor ohms"
	print #1,"Tc = Temperature in degrees C"
	print #1,"Tf = Temperature in degrees F"
	print #1,"Strike any key to return to menu..."

	if inkey$<>"" goto quit		' Quit if key hit
	
	for a%=1 to 32000		' waste some time
	next a%

	if inkey$<>"" then return
	goto ShowTempLoop		' go back through loop
